//
// A2DP test program for OS X version 0.1 programmed by bkc
// This program is using a part of a2play, so,this also is provided under GPL as well as a2play.
// Please see document for detail. 
//

#include <stdio.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOBluetoothUI/IOBluetoothUIUserLib.h>

#include <sys/time.h>

#include "a2dp.h"

static int calc_frame_len(struct sbc_frame_header *hdr)
{
	int tmp, nrof_subbands, nrof_blocks;

	nrof_subbands = (hdr->subbands + 1) * 4;
	nrof_blocks = (hdr->blocks + 1) * 4;

	switch (hdr->channel_mode) {
	case 0x00:
		nrof_subbands /= 2;
		tmp = nrof_blocks * hdr->bitpool;
		break;
	case 0x01:
		tmp = nrof_blocks * hdr->bitpool * 2;
		break;
	case 0x02:
		tmp = nrof_blocks * hdr->bitpool;
		break;
	case 0x03:
		tmp = nrof_blocks * hdr->bitpool + nrof_subbands;
		break;
	default:
		return 0;
	}

	return (nrof_subbands + ((tmp + 7) / 8));
}

static double calc_bit_rate(struct sbc_frame_header *hdr)
{
	int nrof_subbands, nrof_blocks;
	double f;

	nrof_subbands = (hdr->subbands + 1) * 4;
	nrof_blocks = (hdr->blocks + 1) * 4;

	switch (hdr->sampling_frequency) {
	case 0:
		f = 16;
		break;
	case 1:
		f = 32;
		break;
	case 2:
		f = 44.1;
		break;
	case 3:
		f = 48;
		break;
	default:
		return 0;
	}

	return ((8 * (calc_frame_len(hdr) + 4) * f) / (nrof_subbands * nrof_blocks));
}

// Prepare packet headers
static void init_request(struct avdtp_header * header, int request_id)
{
	static int transaction = 0;

	header->packet_type = PACKET_TYPE_SINGLE;
	header->message_type = MESSAGE_TYPE_COMMAND;
	header->transaction_label = transaction;
	header->signal_id = request_id;

	// clear rfa bits
	header->rfa0 = 0;

	transaction = (transaction + 1) & 0xf;
}


///  L2CAP callbacks

#define MAX_DATALEN 0x8000

unsigned char tempData[MAX_DATALEN];
int gIDLLen = 0;

// Event listener
void l2capEL(IOBluetoothL2CAPChannelRef l2capChannel, void *refCon, IOBluetoothL2CAPChannelEvent *event)
{
}

// Incoming data listener
void l2capIDL(IOBluetoothL2CAPChannelRef l2capChannel, void *data,UInt16 len, void *refcon)
{
	if (gIDLLen + len > MAX_DATALEN)
		return; // fatal error

	memcpy(tempData + gIDLLen,data,len);
	gIDLLen += len;
}

void l2capSTRMIDL(IOBluetoothL2CAPChannelRef l2capChannel, void *data,UInt16 len, void *refcon)
{
	fprintf(stderr,"STRMIDL len : %d\n",len);
}


int l2recv(void *dest,int len)
{
	CFRunLoopRunInMode( kCFRunLoopDefaultMode, 1, false );

	if (len > gIDLLen) len = gIDLLen;

	memcpy(dest , tempData , len);

	if ((gIDLLen - len) > 0)	
		memmove(tempData,tempData + len,gIDLLen - len);

	gIDLLen -= len;
	
	return len;
}


int proc_seid(IOBluetoothL2CAPChannelRef ctrlChan, struct acp_seid_info * get_seid_resp,struct sbc_frame_header *hdr)
{
	int ioerr;
	struct getcap_req      put_req;
	struct getcap_resp     cap_resp;
	struct set_config      s_config;
	struct set_config_resp s_resp;
	struct stream_cmd      open_stream;
	struct open_stream_rsp open_resp;
	
	int seid;
	int size;
	int v;
	
	seid = get_seid_resp->acp_seid;
	
	printf("SEID = %d\n",seid);
	
	memset(&put_req, 0, sizeof(put_req));
	init_request(&put_req.header, AVDTP_GET_CAPABILITIES);
	put_req.acp_seid = seid;
	
	ioerr = IOBluetoothL2CAPChannelWriteSync(ctrlChan,&put_req,sizeof(put_req));
	if (ioerr != kIOReturnSuccess)
	{
		printf("get_capabilities  : ng\n");
		return -1;
	}
	
	size = l2recv(&cap_resp,sizeof(cap_resp));
	
	if (size < sizeof(cap_resp) ||
		cap_resp.header.message_type == MESSAGE_TYPE_REJECT ||
		cap_resp.media_type != AUDIO_MEDIA_TYPE)
	{
		printf("not codec param\n");
		return -1;
	}
	
	if (cap_resp.media_codec_type != SBC_MEDIA_CODEC_TYPE)
	{
		if(cap_resp.media_codec_type == MPEG12_MEDIA_CODEC_TYPE)
		{
			printf("SEID %d : MPEG codec\n",seid);
		
		} else
		{
			printf("unknown codec\n");
		}
		return -1;
	}
	
	
	// SBC/MPEG configure
	// fill only 1bit in each field

	
	memset(&s_config, 0, sizeof(s_config));
	init_request(&s_config.header, AVDTP_SET_CONFIGURATION);
	s_config.serv_cap = MEDIA_TRANSPORT_CATEGORY;
	s_config.acp_seid = seid;
	s_config.int_seid = 1;	// how should I choose the int_seid??
	s_config.cap_type = MEDIA_CODEC;
	s_config.length = 6;
	s_config.media_type = AUDIO_MEDIA_TYPE;
	s_config.media_codec_type = SBC_MEDIA_CODEC_TYPE;
	
	switch(hdr->channel_mode)
	{
	case 0:
		v = 8;
	break;
	case 1:
		v = 4;
	break;
	case 2:
		v = 2;
	break;
	case 3:
		v = 1;
	break;
	}

//	printf("chan : %d\n",v);
	s_config.codec_elements.sbc_elements.channel_mode = v;
	
	switch(hdr->sampling_frequency)
	{
		case 0:
			v = 8;
			break;
		case 1:
			v = 4;
			break;
		case 3:
			v = 1;
			break;
		case 2:
		default:
			v = 2; 
			break;
	}
	
//	printf("freq : %d\n",v);

	s_config.codec_elements.sbc_elements.frequency = v; // 44k
	s_config.codec_elements.sbc_elements.allocation_method = 1 << 1;

	switch(hdr->subbands)
	{
	case 0:
		v = 2;
		break;
	case 1:
	default:
		v = 1;
		break;
	}

//	printf("subband : %d\n",v);

	s_config.codec_elements.sbc_elements.subbands = v; // 8band

	switch(hdr->blocks)
	{
	case 0:
		v = 8;
		break;
	case 1:
		v = 4;
		break;
	case 2:
		v = 2;
		break;
	case 3:
	default:
		v = 1;
		break;
	}


//	printf("blocks : %d\n",v);
	s_config.codec_elements.sbc_elements.block_length = v;
	
	s_config.codec_elements.sbc_elements.min_bitpool = cap_resp.codec_elements.sbc_elements.min_bitpool;
	s_config.codec_elements.sbc_elements.max_bitpool = cap_resp.codec_elements.sbc_elements.max_bitpool;

	if (!(cap_resp.codec_elements.sbc_elements.channel_mode & s_config.codec_elements.sbc_elements.channel_mode))
		fprintf(stderr, "headset does not support this channel mode\n");

	if (!(cap_resp.codec_elements.sbc_elements.frequency & s_config.codec_elements.sbc_elements.frequency))
		fprintf(stderr, "headset does not support this frequency\n");

	if (!(cap_resp.codec_elements.sbc_elements.allocation_method & s_config.codec_elements.sbc_elements.allocation_method))
		fprintf(stderr, "headset does not support this allocation_method\n");

	if (!(cap_resp.codec_elements.sbc_elements.subbands & s_config.codec_elements.sbc_elements.subbands))
		fprintf(stderr, "headset does not support this subbands setting\n");


	ioerr = IOBluetoothL2CAPChannelWriteSync(ctrlChan,&s_config,sizeof(s_config));
	if (ioerr != kIOReturnSuccess)
	{
		fprintf(stderr, "couldn't set config seid = %d\n", seid);
		return (-1);
	}

	fprintf(stderr, "Sent set configurations command\n");
	
	size = l2recv(&s_resp,sizeof(s_resp));
	if (size == sizeof(s_resp) - 2)
		printf("set_config : accepted\n");
	else
		printf("set_config : rejected");


	memset(&open_stream, 0, sizeof(open_stream));
	init_request(&open_stream.header, AVDTP_OPEN);
	open_stream.acp_seid = seid;
	
	ioerr = IOBluetoothL2CAPChannelWriteSync(ctrlChan,&open_stream,sizeof(open_stream));
	if (ioerr != kIOReturnSuccess)
	{
		printf("AVDTP_OPEN : ng\n");
		return -1;
	}
	
	size = l2recv(&open_resp,sizeof(open_resp));
	
	if (size < sizeof(open_resp) -1  ||
		open_resp.header.message_type == MESSAGE_TYPE_REJECT)
	{
		printf("OPEN : rejected\n");
		return -1;
	}
	
	return 0;
}

int connectBTdevice(IOBluetoothL2CAPChannelRef *chan,const char *addr,int psm,
	IOBluetoothL2CAPChannelIncomingDataListener listener)
{
	IOReturn ioerr;
	CFStringRef cfaddr;
	IOBluetoothDeviceRef   bdev;
	BluetoothDeviceAddress bdaddr;

	cfaddr = CFStringCreateWithCString(NULL,addr,kCFStringEncodingASCII);
	ioerr = IOBluetoothNSStringToDeviceAddress(cfaddr,&bdaddr);
	CFRelease(cfaddr);
	
	bdev = IOBluetoothDeviceCreateWithAddress(&bdaddr);

	
	if (IOBluetoothDeviceOpenConnection(bdev,NULL,NULL) != kIOReturnSuccess)
	{
		printf("Failed to open connection\n");
		return -1;
	}

	ioerr = IOBluetoothDeviceOpenL2CAPChannelSync(bdev,chan,psm,l2capEL,NULL);
	if (ioerr != kIOReturnSuccess)
	{
		printf("open L2CAP Channel: ng\n");
		IOBluetoothObjectRelease(bdev);
		return -1;
	}

	ioerr = IOBluetoothL2CAPChannelRegisterIncomingDataListener(*chan,listener,NULL);
	if (ioerr != kIOReturnSuccess)
	{
		printf("Failed to register IDL\n");
		IOBluetoothObjectRelease(bdev);
		return -1;
	}
	IOBluetoothObjectRelease(bdev);

	return 0;
	
}

int AVDTPdiscover(IOBluetoothL2CAPChannelRef chan,int *pseid,struct sbc_frame_header *hdr)
{
	IOReturn ioerr;
	int i,size;
	int seid;
	struct sepd_req put_req;
	struct sepd_resp get_resp;

	int last_seid_index;


	*pseid = -1;
	
	memset(&put_req, 0, sizeof(put_req));
	init_request(&put_req.header, AVDTP_DISCOVER);
	
	ioerr = IOBluetoothL2CAPChannelWriteSync(chan,&put_req,sizeof(put_req));
	
	if (ioerr != kIOReturnSuccess)
		return -1;

	size = l2recv(&get_resp,2);
	if (get_resp.header.message_type == 0 && get_resp.header.packet_type == 0)
	{
		printf("discover : try again\n");
		size = l2recv(&get_resp,sizeof(get_resp));
	}
	else
	{
		size = l2recv(((char *)&get_resp)+2,sizeof(get_resp)-2);
		size += 2;
		printf("discover : received \n");
	}

	if (size < sizeof(get_resp) - MAX_ADDITIONAL_CODEC_OCTETS)
	{
		printf("Too short received data!\n");
		return 0;
	}
	
	
	seid = -1;
	last_seid_index = MAX_ADDITIONAL_CODEC - ((sizeof(get_resp)-size)/sizeof(struct acp_seid_info));
	for(i=last_seid_index; i>=0; i--)
	{
		if (proc_seid(chan,&get_resp.infos[i],hdr) == 0)
		{
			seid = get_resp.infos[i].acp_seid;
			break;
		}
	}
	
	*pseid = seid;

		
	return 0;
}

int g_break = 0;

void catch_sigint(int sig)
{
	g_break = 1;
	printf("break\n");
}


long usec_time(void)
{
	struct timeval tv;
	if (gettimeofday(&tv,NULL) < 0)
		return -1;
		
	

	return (tv.tv_sec * 1000000) + tv.tv_usec;
}

long sec_time(void)
{
	struct timeval tv;
	if (gettimeofday(&tv,NULL) < 0)
		return -1;
		
	

	return tv.tv_sec;
}




int btmain(int argc,char *argv[])
{

	int i;
	int size;
	int seid;
	
	FILE *fp;
	
	int frm_len;
	
	IOReturn ioerr;
	int err;
	
	int bytes_in_sec;
	int count_bytes;
	int sleep_period;
	int sleep_usec;
	int current_clock;
	
	int start_time;

	IOBluetoothL2CAPChannelRef ctrlChan;
	IOBluetoothL2CAPChannelRef strmChan;

	struct stream_cmd close_stream;
	struct close_stream_rsp close_resp;
	
	struct stream_cmd start_stream;
	struct start_stream_rsp start_resp;
	
	
	struct media_packet_header packet_header;
	struct media_payload_header payload_header;
	
	int seq_num;
	time_t timestamp;

	unsigned char sendBuf[0x1000];
	int sendBufPtr;

	
	struct sbc_frame_header hdr;

	printf("Bluetooth SBC player ver 0.1\n");
	if (argc < 3)
	{
		printf("Usage <addr> <file>\n");
		return 0;
	}
	
	if (!strcmp(argv[2],"-"))
		fp = stdin;
	else 
		fp = fopen(argv[2],"rb");
		
	if (!fp)
	{
		printf("File open error\n");
		return 1;
	}
	fread(&hdr,sizeof(hdr),1,fp);
	
	if (hdr.syncword != 0x9c)
	{
		printf("This is not an SBC file\n");
		fclose(fp);
		return 1;
	}
	
	frm_len = calc_frame_len(&hdr);
	
	printf("frm_len = %d\n",frm_len);
	
	fseek(fp,0,SEEK_SET);
	
	printf("open : control channnel\n");
	
	if (connectBTdevice(&ctrlChan,argv[1],25,l2capIDL) < 0) // psm is fixed at 25
	{
		printf("connectBTdevice : ng\n");
		return 1;
	}
	
	if (AVDTPdiscover(ctrlChan,&seid,&hdr) < 0)
	{
		printf("discover : ng\n");
		return 1;
	}
	
	
	if (seid == -1)
	{
		printf("Can't open stream port\n");
		return -1;
	}

	// ------ open stream port----
	// 
	
	printf("open : stream channnel\n");
	if (connectBTdevice(&strmChan,argv[1],25,l2capSTRMIDL) < 0) // psm for stream is 25 too.
	{
		printf("connectBTdevice : ng\n");
		return 1;
	}
	
	printf("open_stream : ok\n");

	
	memset(&start_stream, 0, sizeof(start_stream));
	init_request(&start_stream.header, AVDTP_START);
	start_stream.acp_seid = seid;

	ioerr = IOBluetoothL2CAPChannelWriteSync(ctrlChan,&start_stream,sizeof(start_stream));
	
	if (ioerr != kIOReturnSuccess)
	{
		printf("start_stream : ng\n");
	} 
	
	printf("start_stream : ok\n");
		
	size = l2recv(&start_resp,sizeof(start_resp));
	
	if (size < sizeof(start_resp) -2 || 
		start_resp.header.message_type == MESSAGE_TYPE_REJECT)
	{
		printf("start : rejected\n");
		return -1;
	}

	timestamp = 0;
	seq_num = 0;
	
	bytes_in_sec = (calc_bit_rate(&hdr)*1000) / 8;
	
	printf("%d bytes/sec\n",bytes_in_sec);
	sleep_period = bytes_in_sec / 16;
	sleep_usec = (1000000)/16;
	
	printf("sleep_usec : %d\n",sleep_usec);
	start_time = sec_time();
	
	
	int mtu;
	
	mtu = IOBluetoothL2CAPChannelGetOutgoingMTU(strmChan);

	printf("mtu = %d\n",mtu);
	
	signal(SIGINT,catch_sigint);
	
	memset(&payload_header, 0, sizeof(payload_header));
	
	count_bytes = 0;
	
	current_clock = usec_time();
	do {
		memset(&packet_header, 0, sizeof(packet_header));
		packet_header.v = 2;
		packet_header.pt = 1;
		packet_header.sequence_number = htons(seq_num);
		packet_header.timestamp = htonl(timestamp);
		packet_header.ssrc = htonl(1);

		timestamp += (hdr.blocks + 1)*4 * (hdr.subbands + 1)*4;
		
		memcpy(sendBuf, &packet_header, sizeof(packet_header));
		sendBufPtr = sizeof(packet_header);

		payload_header.frame_count = 0;

		// NONSPEC
		
/*		sendBuf[12] = 0xff;
		sendBuf[13] = 0xff;
		
		fread(sendBuf+ MEDIA_PACKET_HEADER_LENGTH,frm_len,1,fp);
*/
		sendBufPtr += sizeof(payload_header);

		do {
		payload_header.frame_count++;
		
		fread(sendBuf + sendBufPtr,frm_len,1,fp);
		count_bytes += frm_len;
		sendBufPtr += frm_len;
		
		}while(!g_break && !feof(fp) && sendBufPtr + frm_len < mtu);
		
		memcpy(sendBuf+sizeof(packet_header), &payload_header, sizeof(payload_header));

//		ioerr = IOBluetoothL2CAPChannelWriteSync(strmChan,sendBuf,sendBufPtr);
		
		ioerr = IOBluetoothL2CAPChannelWriteAsync(strmChan,sendBuf,sendBufPtr,NULL);
		if (ioerr != kIOReturnSuccess)
			printf("sendbuf : ng\n");

		if (count_bytes > sleep_period)
		{
			int i,j;
			j = usec_time();
			i = (current_clock + sleep_usec) - j;
			
			if (i > sleep_usec)
			{
				printf("%ld : too big : %d , current:%d\n",sec_time() - start_time,i,j);
				i = sleep_usec;
			}

			if (i < 0)
			{
				printf("%ld : delayed = %d\n",sec_time() - start_time ,i);
				current_clock = usec_time();
			} else
			{
				usleep(i);
				current_clock = usec_time();
			}
			count_bytes -= sleep_period;
			
		}
		
		seq_num++;

		
	}while(!g_break && !feof(fp));

	
	// ------- close ---------
	memset(&close_stream, 0, sizeof(close_stream));
	memset(&close_resp, 0, sizeof(close_resp));

	// the stream-close used to make the iTech headset lock up and require it to be powercycled
	// should be tested again now that we drain the queue properly

	init_request(&close_stream.header, AVDTP_CLOSE);
	close_stream.acp_seid = seid;
	
	
	ioerr = IOBluetoothL2CAPChannelWriteSync(ctrlChan,&close_stream,sizeof(close_stream));
	
	if (ioerr != kIOReturnSuccess)
		printf("close_stream : ng\n");
	else
		printf("close_stream : ok\n");
		
	size = l2recv(&close_resp,sizeof(close_resp));
	
	if (size < sizeof(close_resp) -1 || 
		close_resp.header.message_type == MESSAGE_TYPE_REJECT)
	{
		printf("close_stream : rejected\n");
	}

	printf("terminate\n");

	IOBluetoothL2CAPChannelCloseChannel(strmChan);
	IOBluetoothL2CAPChannelCloseChannel(ctrlChan);
	
	fclose(fp);

	return 0;
}
